package com.dali.galery

import android.content.Context
import android.graphics.drawable.Drawable
import android.view.View
import com.dali.android.processor.ResourceProperties
import com.dali.galery.network.DaliOrchestrator
import com.dali.galery.network.Result
import com.dali.galery.reflection.DaliClient
import java.lang.ref.WeakReference

/**
 * Image downloading, transformation, and caching manager.
 * Use with(Context) for the global singleton instance.
 */
class Dali {

    fun addView(resourceProperties: ResourceProperties?, view: View?): Boolean {
        if (resourceProperties != null && view != null) {
            var viewList: ArrayList<WeakReference<View>>? = viewMap[resourceProperties]
            if (viewList == null) {
                viewList = ArrayList()
                viewMap[resourceProperties] = viewList
            }
            viewList.add(WeakReference(view))
            return true
        }
        return false
    }

    /**
     * Remove list [View] from waiting list for download [Map]
     */
    fun removeScope(resourceProperties: ResourceProperties) {
        viewMap.remove(resourceProperties)
        if (viewMap.isEmpty()) {
            for ((key, value) in downloadMap) {
                val listener: OnDownloadListener? = value.get()
                if (listener != null) {
                    listener.onDownload()
                } else {
                    downloadMap.remove(key)
                }
            }
        }
    }

    /**
     * Get list [View] by [ResourceProperties]
     */
    fun getScope(resourceProperties: ResourceProperties): ArrayList<WeakReference<View>>? = viewMap[resourceProperties]

    /**
     * Get instance [DaliClient]
     */
    fun getDaliClient(): DaliClient = DaliClient.get()

    /**
     * Network image download listener
     */
    fun interface OnDownloadListener {
        fun onDownload()
    }

    /**
     * Network image download callback
     *
     * Called every time a successful download from the network or cache occurs.
     * For xml markup only
     */
    abstract class Callback {
        abstract fun onSuccess(result: Result.Success<Drawable?>)
        abstract fun onError(error: Result.Error)
    }

    /**
     * Image transformer
     *
     * The transformer changes the image before displaying, and does not affect the cache
     * For xml markup only
     */
    fun interface Transformer {
        fun onTransform(drawable: Drawable, name: String): Drawable
    }

    companion object {

        private val instance = Dali()
        private val viewMap: HashMap<ResourceProperties, ArrayList<WeakReference<View>>> = HashMap()
        private val downloadMap: HashMap<Int, WeakReference<OnDownloadListener>> = HashMap()

        /**
         * Get instance [Dali]
         */
        fun getInstance(): Dali {
            return instance
        }

        /**
         * Add Listener for the image loading event
         * Called after displaying images loaded from a file or from the network
         */
        fun addOnDownloadListener(listener: OnDownloadListener) {
            downloadMap[listener.hashCode()] = WeakReference(listener)
        }

        /**
         * Unsubscribe [addOnDownloadListener]
         */
        fun removeOnDownloadListener(listener: OnDownloadListener?) {
            listener?.let {
                downloadMap.remove(listener.hashCode())
            }
        }

        /**
         * The global default {@link Dali} instance.
         * <p>
         * This instance is automatically initialized with defaults that are suitable to most
         * implementations.
         *
         * You can create your own settings above config with {@link Picasso.RequestCreator}
         *
         * @param context ApplicationContext for get resources.
         * @return RequestCreator
         */
        @JvmStatic
        fun with(context: Context): RequestCreator {
            return RequestCreator(context, getInstance(), DaliOrchestrator.instance)
        }
    }
}