package com.dali.galery

import android.content.Context
import android.graphics.drawable.Drawable
import android.view.View
import androidx.core.content.ContextCompat
import com.dali.galery.network.DaliOrchestrator
import com.dali.android.processor.ResourceProperties

/**
 * Fluent API for building an image download request.
 */
class RequestCreator(
    private val context: Context,
    private val dali: Dali,
    private val daliOrchestrator: DaliOrchestrator,
) {

    private var view: View? = null
    private var resource: ResourceProperties? = null
    private var needPreview: Boolean = false
    private var placeholderDrawable: Drawable? = null
    private var callback: Dali.Callback? = null
    private var transform: Dali.Transformer? = null

    /**
     * Start an image request using the specified path. This is a convenience method for calling load(Uri).
     * Loading an image by a [ResourceProperties]
     */
    fun load(resourceProperties: ResourceProperties?): RequestCreator {
        this.resource = resourceProperties
        return this
    }

    /**
     * Asynchronously fulfills the request into the specified ImageView.
     * Note: This method keeps a weak reference to the ImageView instance and will automatically support object recycling.
     */
    fun into(view: View): RequestCreator {
        this.view = view
        return this
    }

    /**
     * Network download report
     */
    fun callback(callback: Dali.Callback): RequestCreator {
        this.callback = callback
        return this
    }

    /**
     * Transforming an image after downloading from the web
     */
    fun transform(transform: Dali.Transformer): RequestCreator {
        this.transform = transform
        return this
    }

    /**
     * A placeholder drawable to be used while the image is being loaded. If the requested image is not immediately available in the memory cache then this resource will be set on the target ImageView.
     * If you are not using a placeholder image but want to clear an existing image (such as when used in an adapter), pass in null.
     */
    fun placeholder(placeholderDrawable: Int): RequestCreator {
        this.placeholderDrawable = ContextCompat.getDrawable(context, placeholderDrawable)
        return this
    }

    /**
     * A placeholder drawable to be used while the image is being loaded. If the requested image is not immediately available in the memory cache then this resource will be set on the target ImageView.
     * If you are not using a placeholder image but want to clear an existing image (such as when used in an adapter), pass in null.
     */
    fun placeholder(placeholderDrawable: Drawable): RequestCreator {
        this.placeholderDrawable = placeholderDrawable
        return this
    }

    /**
     * Use backgrounds to download images from the web
     * @param preview Use preview background
     */
    fun usePreview(preview: Boolean): RequestCreator {
        this.needPreview = preview
        return this
    }

    /**
     * Download an image from the network.
     */
    fun execute() {
        view?.let { view -> dali.getDaliClient().viewStrategy.onDraw(view, placeholderDrawable) }
        dali.addView(resource, view)
        val res = resource?.resPreview
        if (needPreview && res != null && res != 0) {
            view?.let { view ->
                dali.getDaliClient().viewStrategy.onDraw(
                    view,
                    ContextCompat.getDrawable(context, res)
                )
            }
        }
        daliOrchestrator.update(resource, context, callback, transform)
    }
}