package com.dali.galery.network

import android.content.Context
import android.graphics.drawable.Drawable
import android.view.View
import androidx.core.content.ContextCompat
import com.dali.android.processor.ResourceProperties
import com.dali.galery.Dali
import com.dali.galery.reflection.DaliClient
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.launch
import kotlinx.coroutines.withContext
import java.lang.ref.WeakReference
import kotlin.reflect.KClass

/**
 * Controls the selection of the image upload source
 */
class DaliOrchestrator {

    private val daliApi = DaliApi()
    private val dali = Dali.getInstance()

    /**
     * Update an image from the waiting list by the key [ResourceProperties]
     */
    fun update(
        resourceProperties: ResourceProperties?,
        context: Context?,
        callback: Dali.Callback?,
        transformer: Dali.Transformer?,
    ) {
        if (context == null || resourceProperties == null) return
        if (!loadListDrawable.contains(resourceProperties)) {
            loadListDrawable.add(resourceProperties)
            CoroutineScope(Dispatchers.IO).launch {
                // load = false то отображаем resPreview
                if (!DaliClient.get().getImageScopeSettingByName(resourceProperties.name).load) {
                    withContext(Dispatchers.Main) {
                        draw(resourceProperties, null, context, transformer)
                        removeLoader(resourceProperties)
                    }
                    return@launch
                }

                // load = true загружаем из сети
                val result = daliApi.loadImage(resourceProperties.uriView, context, resourceProperties.name)
                withContext(Dispatchers.Main) {
                    when (result) {
                        is Result.Success -> {
                            draw(resourceProperties, result.data, context, transformer)
                            removeLoader(resourceProperties)
                            callback?.onSuccess(result)
                        }

                        is Result.Error -> {
                            result.exception.printStackTrace()
                            draw(resourceProperties, null, context, transformer)
                            removeLoader(resourceProperties)
                            callback?.onError(result)
                        }
                    }
                }
            }
        }
    }

    private fun removeLoader(resourceProperties: ResourceProperties) {
        loadListDrawable.remove(resourceProperties)
        dali.removeScope(resourceProperties)
    }

    private fun draw(
        resourceProperties: ResourceProperties,
        drw: Drawable?,
        context: Context,
        transformer: Dali.Transformer?,
    ) {
        var drawable: Drawable? = drw
        if (drawable == null && resourceProperties.resPreview != 0) {
            drawable = ContextCompat.getDrawable(context, resourceProperties.resPreview)
        }
        val daliClient: DaliClient = DaliClient.get()
        val viewWeakReferenceList: ArrayList<WeakReference<View>>? = dali.getScope(resourceProperties)

        if (drawable != null) {
            viewWeakReferenceList?.forEach { viewWeak ->
                viewWeak.get()?.let { view ->
                    if (transformer != null) daliClient.viewStrategy.onDraw(view, transformer.onTransform(drawable, resourceProperties.name))
                    else daliClient.viewStrategy.onDraw(view, drawable)
                }
            }
        }
    }

    companion object {

        /**
         * Remove cache file
         */
        @JvmStatic
        fun removeCacheFile(context: Context, resourceProperties: ResourceProperties?) {
            instance.daliApi.removeCacheFile(context, resourceProperties?.name)
        }

        /**
         * Remove cache files
         */
        @JvmStatic
        fun removeCacheFiles(context: Context, args: Array<ResourceProperties?>) {
            args.forEach { resourceProperties ->
                instance.daliApi.removeCacheFile(context, resourceProperties?.name)
            }
        }

        /**
         * Remove cache scope
         */
        fun removeCacheScope(context: Context, kClass: KClass<*>) {
            val fields = kClass.java.declaredFields
            fields.forEach { field ->
                //TODO убрать replace Res после рефакторинга
                instance.daliApi.removeCacheFile(context, "${kClass.simpleName?.replace("Res", "")}.${field.name}")
            }
        }

        val instance = DaliOrchestrator()
        val loadListDrawable: ArrayList<ResourceProperties> = ArrayList()
    }
}