package com.dali.galery.reflection

import androidx.annotation.MainThread
import com.dali.galery.Dali
import com.dali.galery.ImageScopeSetting
import com.dali.galery.ViewStrategy
import com.dali.galery.reflection.inflate.data.InflateRequest
import com.dali.galery.reflection.inflate.data.InflateResult
import com.dali.galery.reflection.interceptor.InterceptView
import com.dali.galery.reflection.interceptor.Interceptor
import com.dali.galery.reflection.viewcategory.FallbackViewCreationInterceptor
import com.dali.galery.reflection.viewcategory.InterceptorChain
import okhttp3.OkHttpClient
import okhttp3.Request
import kotlin.reflect.KClass

/**
 * Library Dali download pictures from the net
 * You can specify the URL in xml to download images from the web
 */
class DaliClient private constructor(
    /** ImageStrategy.  */
    @get:JvmName("imageStrategy")
    val viewStrategy: ViewStrategy,

    /** Store the resourceId for the layout used to inflate the View in the View tag.  */
    @get:JvmName("isStoreLayoutResId")
    val isStoreLayoutResId: Boolean,

    @get:JvmName("okHttpClient")
    val okHttpClient: OkHttpClient?,

    @get:JvmName("okHttpRequest")
    val okHttpRequest: Request.Builder?,

    /**
     * Base URL for loading pictures
     * Base URL for downloading pictures
     *
     * uri get from annotations @UriView
     * requestUrl = baseURL + uri
     * Example:
     *
     * baseURL https://example.com/
     * uri cat/cat_PNG50534.png
     * requestUrl https://example.com/cat/cat_PNG50534.png
     *
     */
    @get:JvmName("baseUrl")
    var baseUrl: String,

    @get:JvmName("statusCodes")
    val responseStatusCodes: IntRange,

    var callback: Dali.Callback? = null,

    var transformer: Dali.Transformer? = null,

    private val imageScopeSettings: HashMap<String, ImageScopeSetting> = HashMap(),
) {

    /** List of interceptors. */
    private val interceptorsWithFallback = mutableListOf<Interceptor>()

    init {
        interceptorsWithFallback.add(InterceptView())
        interceptorsWithFallback.add(FallbackViewCreationInterceptor())
    }

    /**
     * Add custom parameters to the scope
     * Applies to scope and overrides Default
     *
     * @param kClass class scope
     * @param additionalUri additionalUri
     * @param dirName
     *
     * uri get from annotations @UriView
     * requestUrl = baseURL + additionalUri + uri
     * Example:
     *
     * baseURL https://example.com/
     * additionalUri cat/
     * uri cat_PNG50534.png
     *
     * requestUrl https://example.com/cat/cat_PNG50534.png
     *
     * !!Attention!!
     * URL without validation
     */
    fun addImageScopeSettings(kClass: KClass<*>, additionalUri: String = "", dirName: String = "") {
        imageScopeSettings[kClass.simpleName.toString()] = ImageScopeSetting(additionalUri, dirName)
    }

    /**
     * Add custom parameters to the scope
     * Applies to scope and overrides Default
     *
     * uri get from annotations @UriView
     * requestUrl = baseURL + additionalUri + uri
     * Example:
     *
     * baseURL https://example.com/
     * additionalUri cat/
     * uri cat_PNG50534.png
     *
     * requestUrl https://example.com/cat/cat_PNG50534.png
     *
     * !!Attention!!
     * URL without validation
     */
    fun addImageScopeSettingsDefault(additionalUri: String = "", dirName: String = "") {
        imageScopeSettings[DEFAULT] = ImageScopeSetting(uri = additionalUri, dirName = dirName)
    }

    /**
     * Add custom options for disable load
     *
     * !!Attention!!
     * Option to disable downloading from the network and uses resPreview
     */
    fun addImageScopeSettingsRemote(kClass: KClass<*>) {
        imageScopeSettings[kClass.simpleName.toString()] = ImageScopeSetting(load = false)
    }

    internal fun inflate(originalRequest: InflateRequest): InflateResult {
        val chain = InterceptorChain(
            interceptorsWithFallback, 0,
            originalRequest
        )
        return chain.proceed(originalRequest)
    }

    fun getImageScopeSettingByName(name: String): ImageScopeSetting {
        return imageScopeSettings[name.split(".")[0]] ?: return imageScopeSettings[DEFAULT] ?: ImageScopeSetting()
    }

    class Builder internal constructor() {

        /** ImageStrategy. */
        private var viewStrategy: ViewStrategy? = null

        /** Store the resourceId for the layout used to inflate the View in the View tag. */
        private var storeLayoutResId = false

        /** Save an OkHttpClient to request images from the web. */
        private var okHttpClient: OkHttpClient? = null

        /** Save a Request to request images from the web. */
        private var okHttpRequest: Request.Builder? = null

        /** Save a Request Url to request images from the web. */
        private var baseUrl: String = ""

        /** Save a response status codes for the web request. */
        private var responseStatusCodes: IntRange = 200..299

        /** Save a callback*/
        private var callback: Dali.Callback? = null

        /** Save a transformer */
        private var transformer: Dali.Transformer? = null

        /**
         * Adding strategies for displaying pictures
         *
         * Optional parameter
         * If necessary, default is used
         *
         * @param strategy [ViewStrategy]
         */
        fun imageStrategy(strategy: ViewStrategy) = apply {
            viewStrategy = strategy
        }

        /**
         * Adding OkHttpClient for loading pictures
         *
         * Optional parameter
         * If necessary, default is used
         *
         * @param client  [ViewStrategy]
         */
        fun okHttpClient(client: OkHttpClient) = apply {
            okHttpClient = client
        }

        /**
         * Adding OkHttp Request for loading pictures
         *
         * Optional parameter
         * If necessary, default is used
         *
         * @param request [Request]
         */
        fun request(request: Request.Builder) = apply {
            okHttpRequest = request
        }

        /**
         * Adding a base URL for loading pictures
         * Base URL for downloading pictures
         * Can be changed later
         *
         * uri get from annotations @UriView
         * requestUrl = baseURL + uri
         * Example:
         * baseURL https://example.com/
         * uri cat/cat_PNG50534.png
         * requestUrl https://example.com/cat/cat_PNG50534.png
         *
         * @param url
         */
        fun baseUrl(url: String) = apply {
            baseUrl = url
        }

        /**
         * Range of successful response codes
         *
         * Optional parameter
         * If necessary, default is used
         *
         * Replaces fun successResponseStatusCode
         * default 200..299
         *
         * @param codes
         */
        fun successResponseStatusCodes(codes: IntRange) = apply {
            responseStatusCodes = codes
        }

        /**
         * Successful response code
         *
         * Optional parameter
         * If necessary, default is used
         *
         * Replaces fun successResponseStatusCodes
         * default 200..299
         *
         * @param code
         */
        fun successResponseStatusCode(code: Int) = apply {
            responseStatusCodes = code..code
        }

        /**
         * Network download report
         * Called every time a successful download from the network or cache occurs.
         *
         * For xml markup only
         */
        fun callback(callback: Dali.Callback?) = apply {
            this.callback = callback
        }

        /**
         * Transforming an image after downloading
         * The transformer changes the image before displaying, and does not affect the cache
         *
         * For xml markup only
         */
        fun transform(transformer: Dali.Transformer) = apply {
            this.transformer = transformer
        }

        fun build(): DaliClient = DaliClient(
            viewStrategy = viewStrategy ?: ViewStrategyDefault(),
            isStoreLayoutResId = storeLayoutResId,
            okHttpClient = okHttpClient,
            okHttpRequest = okHttpRequest,
            baseUrl = baseUrl,
            responseStatusCodes = responseStatusCodes,
            callback = callback,
            transformer = transformer
        )
    }

    companion object {

        private var INSTANCE: DaliClient? = null
        private const val DEFAULT = "DEFAULT"

        @JvmStatic
        fun init(daliView: DaliClient?) {
            INSTANCE = daliView
        }

        @JvmStatic
        @MainThread
        fun get(): DaliClient {
            return INSTANCE ?: builder().build().also { INSTANCE = it }
        }

        @JvmStatic
        fun builder(): Builder {
            return Builder()
        }
    }
}