package com.dali.galery.reflection.viewcategory

import android.content.Context
import android.util.AttributeSet
import android.view.LayoutInflater
import android.view.View
import android.view.ViewGroup
import com.dali.galery.R
import com.dali.galery.reflection.DaliClient
import com.dali.galery.reflection.FallbackViewCreator
import com.dali.galery.reflection.inflate.data.InflateRequest

internal class ViewDaliLayoutInflater(
    original: LayoutInflater,
    newContext: Context,
    cloned: Boolean,
) : LayoutInflater(original, newContext), ViewDaliActivityFactory {

    private val nameAndAttrsViewCreator: FallbackViewCreator = NameAndAttrsViewCreator(this)
    private val parentAndNameAndAttrsViewCreator: FallbackViewCreator = ParentAndNameAndAttrsViewCreator(this)

    private var storeLayoutResId = DaliClient.get().isStoreLayoutResId

    init {
        setUpLayoutFactories(cloned)
    }

    override fun cloneInContext(newContext: Context): LayoutInflater {
        return ViewDaliLayoutInflater(this, newContext, true)
    }

    override fun inflate(resource: Int, root: ViewGroup?, attachToRoot: Boolean): View? {
        val view = super.inflate(resource, root, attachToRoot)
        if (view != null && storeLayoutResId) {
            view.setTag(R.id.view_dali_layout_res, resource)
        }
        return view
    }

    override fun setFactory(factory: Factory) {
        // Only set our factory and wrap calls to the Factory trying to be set!
        if (factory !is WrapperFactory) {
            super.setFactory(
                WrapperFactory(factory)
            )
        } else {
            super.setFactory(factory)
        }
    }

    override fun setFactory2(factory2: Factory2) {
        // Only set our factory and wrap calls to the Factory2 trying to be set!
        if (factory2 !is WrapperFactory2) {
            // LayoutInflaterCompat.setFactory(this, new WrapperFactory2(factory2, mViewPumpFactory));
            super.setFactory2(
                WrapperFactory2(factory2)
            )
        } else {
            super.setFactory2(factory2)
        }
    }

    /**
     * The Activity onCreateView (PrivateFactory) is the third port of call for LayoutInflation.
     * We opted to manual injection over aggressive reflection, this should be less fragile.
     */
    override fun onActivityCreateView(
        parent: View?,
        view: View,
        name: String,
        context: Context,
        attrs: AttributeSet?,
    ): View? {
        return DaliClient.get()
            .inflate(
                InflateRequest(
                    name = name,
                    context = context,
                    attrs = attrs,
                    parent = parent,
                    fallbackViewCreator = ActivityViewCreator(view)
                )
            )
            .view
    }

    /**
     * The LayoutInflater onCreateView is the fourth port of call for LayoutInflation.
     * BUT only for none CustomViews.
     */
    @Throws(ClassNotFoundException::class)
    override fun onCreateView(parent: View?, name: String, attrs: AttributeSet?): View? {
        return DaliClient.get()
            .inflate(
                InflateRequest(
                    name = name,
                    context = context,
                    attrs = attrs,
                    parent = parent,
                    fallbackViewCreator = parentAndNameAndAttrsViewCreator
                )
            )
            .view
    }

    /**
     * The LayoutInflater onCreateView is the fourth port of call for LayoutInflation.
     * BUT only for none CustomViews.
     * Basically if this method doesn't inflate the View nothing probably will.
     */
    @Throws(ClassNotFoundException::class)
    override fun onCreateView(name: String, attrs: AttributeSet?): View? {
        return DaliClient.get()
            .inflate(
                InflateRequest(
                    name = name,
                    context = context,
                    attrs = attrs,
                    fallbackViewCreator = nameAndAttrsViewCreator
                )
            )
            .view
    }

    /**
     * We don't want to unnecessary create/set our factories if there are none there. We try to be
     * as lazy as possible.
     */
    private fun setUpLayoutFactories(cloned: Boolean) {
        if (cloned) return
        // If we are HC+ we get and set Factory2 otherwise we just wrap Factory1
        if (factory2 != null && factory2 !is WrapperFactory2) {
            // Sets both Factory/Factory2
            factory2 = factory2
        }
        // We can do this as setFactory2 is used for both methods.
        if (factory != null && factory !is WrapperFactory) {
            factory = factory
        }
    }

    private fun superOnCreateView(parent: View?, name: String, attrs: AttributeSet?): View? {
        return try {
            super.onCreateView(parent, name, attrs)
        } catch (e: ClassNotFoundException) {
            null
        }
    }

    private fun superOnCreateView(name: String, attrs: AttributeSet?): View? {
        return try {
            super.onCreateView(name, attrs)
        } catch (e: ClassNotFoundException) {
            null
        }
    }

    private class ActivityViewCreator(
        private val view: View,
    ) : FallbackViewCreator {

        override fun onCreateView(
            parent: View?,
            name: String,
            context: Context,
            attrs: AttributeSet?,
        ): View {
            return view
        }
    }

    private class ParentAndNameAndAttrsViewCreator(
        private val inflater: ViewDaliLayoutInflater,
    ) : FallbackViewCreator {

        override fun onCreateView(
            parent: View?, name: String, context: Context,
            attrs: AttributeSet?,
        ): View? {
            return inflater.superOnCreateView(parent, name, attrs)
        }
    }

    private class NameAndAttrsViewCreator(
        private val inflater: ViewDaliLayoutInflater,
    ) : FallbackViewCreator {

        override fun onCreateView(
            parent: View?,
            name: String,
            context: Context,
            attrs: AttributeSet?,
        ): View? {
            // This mimics the {@code PhoneLayoutInflater} in the way it tries to inflate the base
            // classes, if this fails its pretty certain the app will fail at this point.
            var view: View? = null
            for (prefix in CLASS_PREFIX_LIST) {
                try {
                    view = inflater.createView(name, prefix, attrs)
                    if (view != null) {
                        break
                    }
                } catch (ignored: ClassNotFoundException) {
                }
            }
            // In this case we want to let the base class take a crack
            // at it.
            if (view == null) view = inflater.superOnCreateView(name, attrs)
            return view
        }
    }

    // ===
    // Wrapper Factories
    // ===

    /**
     * Factory 1 is the first port of call for LayoutInflation
     */
    private class WrapperFactory(factory: Factory) : Factory {

        private val viewCreator: FallbackViewCreator = WrapperFactoryViewCreator(factory)

        override fun onCreateView(name: String, context: Context, attrs: AttributeSet): View? {
            return DaliClient.get()
                .inflate(
                    InflateRequest(
                        name = name,
                        context = context,
                        attrs = attrs,
                        fallbackViewCreator = viewCreator
                    )
                )
                .view
        }
    }

    private class WrapperFactoryViewCreator(
        private val factory: Factory,
    ) : FallbackViewCreator {

        override fun onCreateView(
            parent: View?,
            name: String,
            context: Context,
            attrs: AttributeSet?,
        ): View? {
            return factory.onCreateView(name, context, attrs!!)
        }
    }

    /**
     * Factory 2 is the second port of call for LayoutInflation
     */
    private open class WrapperFactory2(factory2: Factory2) : Factory2 {
        private val viewCreator = WrapperFactory2ViewCreator(factory2)
        override fun onCreateView(parent: View?, name: String, context: Context, attrs: AttributeSet): View? {
            return DaliClient.get()
                .inflate(
                    InflateRequest(
                        name = name,
                        context = context,
                        attrs = attrs,
                        parent = parent,
                        fallbackViewCreator = viewCreator
                    )
                )
                .view
        }

        override fun onCreateView(name: String, context: Context, attrs: AttributeSet): View? {
            return onCreateView(null, name, context, attrs)
        }
    }

    private open class WrapperFactory2ViewCreator(
        protected val factory2: Factory2,
    ) : FallbackViewCreator {

        override fun onCreateView(
            parent: View?,
            name: String,
            context: Context,
            attrs: AttributeSet?,
        ): View? {
            return factory2.onCreateView(parent, name, context, attrs!!)
        }
    }

    companion object {
        private val CLASS_PREFIX_LIST = setOf("android.widget.", "android.webkit.")
    }
}
